#include "3dView.h"
#include "GLFoundation.h"
#include "MapSegments.h"
#include "Bitmap.h"
#include "ImageArray.h"
#include <math.h>
#include <stdio.h>

C3DView *SceneViewer;

#define COSSIZE		4096
#define PI			3.14159265359
#define D_PI		6.28318530718
#define H_PI		1.5707963268

#define Cos(v)	CosTable[(int)((v)*COSSIZE/D_PI) & (COSSIZE-1)]
#define Sin(v)	CosTable[(int)((H_PI - (v))*COSSIZE/D_PI) & (COSSIZE-1)]
#define Tan(v)	(Sin(v)/Cos(v))

/* the extra divide for Tan is worth it to save the table size and hence the processor cache */

#define TABLE_QUANTUM	((D_PI)/(float)COSSIZE)

#define MAX_DEPTH	64

C3DView::C3DView()
{
	CosTable = new float[COSSIZE];

	int c = COSSIZE;
	while(c--)
	{
		float ang = (float)c * D_PI / (float)COSSIZE;
		CosTable[c] = cos(ang);
	}

	CastAngles = NULL;
	SizeMultipliers = NULL;
	
	c = 26;
	while(c--)
	{
		RenderPaths[c].SegHead = NULL;
		RenderPaths[c].ObjHead = NULL;
		RenderPaths[c].CornHead = NULL;
	}
}

C3DView::~C3DView()
{
	delete[] CosTable;
	delete[] CastAngles;
	delete[] SizeMultipliers;
	int c = 26;
	while(c--)
	{
		FreeSegments(RenderPaths[c].SegHead);
		FreeObjects(RenderPaths[c].ObjHead);
		FreeCorners(RenderPaths[c].CornHead);
	}
}

void C3DView::SetRenderArea(float x1, float y1, float x2, float y2)
{
	C3DView::x1 = x1; C3DView::x2 = x2;
	C3DView::y1 = y1; C3DView::y2 = y2;
	Steps = GLFoundation.GetNumScanlines(y1, y2);
	StepSize = (y2 - y1) / (float)Steps;

	delete[] CastAngles;
	delete[] SizeMultipliers;
	CastAngles = new float[Steps];
	SizeMultipliers = new float[Steps];
	int c = Steps;
	float YPos, YAdd;
	YAdd = 1.5f / (Steps-1);
	YPos = - 0.75f;

	while(c--)
	{
		CastAngles[c] = atan2(YPos, 1.0f);
		SizeMultipliers[c] = cos(CastAngles[c]);
		YPos += YAdd;
	}

	CentreX = (x1 + x2)*0.5f;
}


void C3DView::EffectCurve(float Curvature, float CurveOffset)
{
	Skew += CurveOffset*SkewMultiplier;

	CurveOffset *= H_PI / MAX_DEPTH;
	Curvature *= 100.0f;

	Skew += Curvature*(Cos(CurveOffset) - 1.0f);
	SkewMultiplier -= (H_PI*Curvature*Sin(CurveOffset)) / MAX_DEPTH;
}

void C3DView::AddScanline(float Distance, float SurfaceY, int path, float x)
{
	/* just do this like a freakin' idiot. A good solution will buffer all these and kill all the GL state changes */
	float Add = (x2-x1) / (2.0f*Distance);
	float Start = CentreX - (4.0f+x)*Add;

	float Offset;
	Offset = fmod(SurfaceY, 1.0f);
	unsigned short *Row = Paths[path].MaterialMap[(int)SurfaceY];

	for(int c = 0; c < 8; c++)
	{
		float LOffset = (Row[c]&TILE_VFLIP) ? Offset : 1.0f - Offset;
		CBitmap *Img = MaterialArray->GetImage(Row[c]&TILE_MASK);
		if(Img)
		{
			Img->Activate();

			glBegin(GL_QUADS);
				if(Row[c]&TILE_HFLIP)
				{
					glTexCoord2f(1, LOffset); glVertex2f(Start, DrawY); glVertex2f(Start, DrawY+StepSize);
					glTexCoord2f(0, LOffset); glVertex2f(Start+Add, DrawY+StepSize); glVertex2f(Start+Add, DrawY);
				}
				else
				{
					glTexCoord2f(0, LOffset); glVertex2f(Start, DrawY); glVertex2f(Start, DrawY+StepSize);
					glTexCoord2f(1, LOffset); glVertex2f(Start+Add, DrawY+StepSize); glVertex2f(Start+Add, DrawY);
				}
			glEnd();
		}
		Start += Add;
	}
}

bool C3DView::DrawPathFloor(int p, float x, float y, float z, RenderSegment *Seg, RenderCorner *CornerInfo, float Curvature)
{
	bool res = true;

	while(Lines < Steps)
	{
		float Angle;
		float RelativeY, RelativeZ;
		float projy;

		/* find first visible segment */
		while(1)
		{
			if(Seg->ArcStart-ArcBase > MAX_DEPTH)
				goto exit; /* MEGA YUCK, but it was either this or nested functions */

			Angle = Seg->Angle - CastAngles[Lines];

			RelativeY = Cos(Seg->Angle)*(z - Seg->CumulativeZ) + Sin(Seg->Angle)*(y - Seg->CumulativeY);
			RelativeZ = - Cos(Seg->Angle)*(y - Seg->CumulativeY) + Sin(Seg->Angle)*(z - Seg->CumulativeZ);

			if(Angle < H_PI-TABLE_QUANTUM && Angle > -H_PI+TABLE_QUANTUM && RelativeZ > 0)
			{
				projy = RelativeZ*Tan(Angle);
				projy += RelativeY + Seg->ArcStart;
				if(projy < Seg->Next->ArcStart) //projy >= Seg->ArcStart && 
					break;
			}

			Seg = Seg->Next;
			Seg->ScrLowY = DrawY+StepSize;

			if(!Seg->Next)
			{
				EffectCurve(Curvature, Seg->ArcStart - CurveBase);
				CurveBase = Seg->ArcStart;
				res = false;
				goto exit;
			}
		}

		/* draw as many scanlines as possible */
		while(Lines < Steps)
		{
			projy = RelativeZ*Tan(Angle);
			float Distance = (projy*Sin(Angle) + RelativeZ*Cos(Angle))*SizeMultipliers[Lines];

			projy += RelativeY + Seg->ArcStart;
			if(projy >= Seg->Next->ArcStart)
				break;

			while(CornerInfo && projy > CornerInfo->y)
			{
				EffectCurve(Curvature, CornerInfo->y - CurveBase);
				CurveBase = CornerInfo->y;
				Curvature = CornerInfo->Curvature;

				CornerInfo->Skew = Skew;
				CornerInfo->SkewMultiplier = SkewMultiplier;
				CornerInfo->CurveBase = CornerInfo->y;

				CornerInfo = CornerInfo->Next;
			}

			if(projy - ArcBase > MAX_DEPTH)
				goto exit;

			float NewSkew = Curvature*100.0f*(Cos(H_PI*(projy - CurveBase)/MAX_DEPTH) - 1.0f);
			NewSkew += (projy - CurveBase)*SkewMultiplier;

			AddScanline(Distance, projy - Paths[p].StartY, p, NewSkew+Skew+x);

			Lines++;
			DrawY -= StepSize;

			Angle = Seg->Angle - CastAngles[Lines];
			if(Angle >= H_PI-TABLE_QUANTUM || Angle <= -H_PI+TABLE_QUANTUM)
				break;
		}
	}
	exit:
	while(CornerInfo && (CornerInfo->y - ArcBase) < MAX_DEPTH)
	{
		EffectCurve(Curvature, CornerInfo->y - CurveBase);
		CurveBase = CornerInfo->CurveBase = CornerInfo->y;
		Curvature = CornerInfo->Curvature;
		CornerInfo->Skew = Skew;
		CornerInfo->SkewMultiplier = SkewMultiplier;

		CornerInfo = CornerInfo->Next;
	}

	return res;
}

void C3DView::DrawPathObjects(int p, float x, float y, float z)
{
	/* add objects */
	RenderObject *Obj = RenderPaths[p].ObjHead;
	while(Obj && Obj->ArcPos > ArcBase+MAX_DEPTH)
		Obj = Obj->Next;
	while(Obj)
	{
		CBitmap *Img = ObjectArray->GetImage(Obj->ImageId&TILE_MASK);
		if(Img)
		{
			GLFoundation.SetScissor(x1, y1, x2, Obj->ParentSegment->ScrLowY);

			float ScrX, ScrY, ScrZ, Width, Height;

			ScrY = Obj->OffsetY + Obj->ParentSegment->CumulativeY - y;
			if(ScrY < 0.1f) return;

			ScrX = Obj->WorldX - x;
			ScrZ = Obj->OffsetZ + Obj->ParentSegment->CumulativeZ - z;

			// apply skew, etc
			if(Obj->ParentCorner)
			{
				float NewSkew = Obj->ParentCorner->Curvature*100.0f*(Cos(H_PI*(Obj->ArcPos - Obj->ParentCorner->CurveBase)/MAX_DEPTH) - 1.0f);
				NewSkew += (Obj->ArcPos - Obj->ParentCorner->CurveBase)*Obj->ParentCorner->SkewMultiplier;
				ScrX -= NewSkew + Obj->ParentCorner->Skew;
			}

			ScrX = (x1+x2)*0.5f + 0.5f*(x1+x2)*ScrX/ScrY;
			ScrZ = (y1+y2)*0.5f - 0.5f*(x1+x2)*ScrZ/ScrY;
			Width = (float)Img->Width() / (256.0f*ScrY);
			Height = (float)Img->Height() / (256.0f*ScrY);

			Img->Draw(ScrX - Width*0.5f, ScrZ - Height, ScrX + Width*0.5f, ScrZ, Obj->ImageId);
		}
		Obj = Obj->Next;
	}
}

void C3DView::DrawPath(int p, float x, float y, float z, RenderSegment *Seg)
{
	/* figure out start segment if not specified */
	if(!Seg)
	{
		Seg = RenderPaths[p].SegHead;
		while(Seg->Next && Seg->Next->CumulativeY < y)
			Seg = Seg->Next;
	}

	/* search forward to relevant corner, with similar stupidity */
	float Curvature = 0;
	RenderCorner *CornerInfo = RenderPaths[p].CornHead;

	if(CornerInfo)
	{
		/* skip past corners */
		while(CornerInfo->Next && CornerInfo->Next->y < ArcBase)
			CornerInfo = CornerInfo->Next;

		/* take info from current corner if it has already occurred */
		if(CornerInfo->y < ArcBase)
		{
			CornerInfo->CurveBase = CurveBase = ArcBase;
			Curvature = CornerInfo->Curvature;
			CornerInfo->Skew = Skew;
			CornerInfo->SkewMultiplier = SkewMultiplier;
			CornerInfo = CornerInfo->Next;
		}
	}

	if(!DrawPathFloor(p, x, y, z, Seg, CornerInfo, Curvature))
	{
		if(Paths[p].Next[0] != -1)
		{
			/* backup important state variables */
			int LLines = Lines;
			float LDrawY = DrawY;
			float LSkew = Skew;
			float LSkewMultiplier = SkewMultiplier;
			float LCurveBase = CurveBase;

			DrawPathFloor(Paths[p].Next[0], x, y, z, RenderPaths[Paths[p].Next[0]].SegHead, RenderPaths[Paths[p].Next[0]].CornHead, 0);

			/* restore important state variables */
			Lines = LLines;
			DrawY = LDrawY;
			Skew = LSkew;
			SkewMultiplier = LSkewMultiplier;
			CurveBase = LCurveBase;
		}
		if(Paths[p].Next[1] != -1)
		{
			DrawPathFloor(Paths[p].Next[1], x, y, z, RenderPaths[Paths[p].Next[1]].SegHead, RenderPaths[Paths[p].Next[1]].CornHead, 0);
			DrawPathObjects(Paths[p].Next[1], x, y, z);
		}
		if(Paths[p].Next[0] != -1)
			DrawPathObjects(Paths[p].Next[0], x, y, z);
	}
	DrawPathObjects(p, x, y, z);
}

void C3DView::DrawTrackPOV(int p, float x, float y, float z)
{
	/* convert from track coordinates, then call Draw */
	RenderSegment *Seg = RenderPaths[p].SegHead;
	while(Seg->Next && Seg->Next->ArcStart < y)
		Seg = Seg->Next;

	float WorldY, WorldZ, Ratio;
	Ratio = (y - Seg->ArcStart) / (Seg->Next->ArcStart - Seg->ArcStart);
	WorldY = Seg->CumulativeY + Ratio*Seg->y;
	WorldZ = Seg->CumulativeZ + Ratio*Seg->z + z;

	Draw(p, x, WorldY, WorldZ);
}

void C3DView::Draw(int p, float x, float y, float z)
{
	/* set viewport, enable texturing */
	GLFoundation.SetScissor(x1, y1, x2, y2);

	/* find first potentially visible segment - stupidly */
	RenderSegment *Seg = RenderPaths[p].SegHead;
	while(Seg->Next && Seg->Next->CumulativeY < y)
		Seg = Seg->Next;
	Seg->ScrLowY = y2;
	if(!Seg->Next) return;

	/* move up/down to track, determine base arc pos */
	float Ratio = (y - Seg->CumulativeY) / Seg->y;
	ArcBase = Ratio * (Seg->Next->ArcStart - Seg->ArcStart) + Seg->ArcStart;

	Lines = 0;
	DrawY = y2;

	/* consider drawing brother if relevant */
	if(((int)ArcBase%SEGMENT_LENGTH) < MAX_DEPTH && Paths[p].Brother != -1)
	{
		float Sk = GetSkew(Paths[p].Brother, Paths[p].StartY, ArcBase);
		float SkM = SkewMultiplier;
		Sk -= GetSkew(p, Paths[p].StartY, ArcBase);

		Skew = Sk;
		SkewMultiplier = SkM - SkewMultiplier;
		CurveBase = ArcBase;
		DrawPath(Paths[p].Brother, x, y, z, NULL);

		Lines = 0;
		DrawY = y2;
	}

	GLFoundation.SetScissor(x1, y1, x2, y2);
	Skew = 0;
	SkewMultiplier = 0;
	CurveBase = ArcBase;
	DrawPath(p, x, y, z, Seg);

	GLFoundation.DisableScissor();
}

float C3DView::GetSkew(int path, float starty, float endy)
{
	float Curvature = 0;
	RenderCorner *CornerInfo = RenderPaths[path].CornHead;
	Skew = SkewMultiplier = 0;

	if(CornerInfo)
	{
		/* skip past corners */
		while(CornerInfo->Next && CornerInfo->Next->y < starty)
			CornerInfo = CornerInfo->Next;

		/* take info from current corner if it has already occurred */
		if(CornerInfo->y < starty)
		{
			CurveBase = starty;
			Curvature = CornerInfo->Curvature;
			CornerInfo = CornerInfo->Next;
		}
	}

	/* do all corners in range */
	while(CornerInfo && CornerInfo->y < endy)
	{
		EffectCurve(Curvature, CornerInfo->y - CurveBase);
		CurveBase = CornerInfo->y;
		Curvature = CornerInfo->Curvature;
		CornerInfo = CornerInfo->Next;
	}
	
	/* now do a little more, because endy is probably after CornerInfo->y */
	EffectCurve(Curvature, endy - CurveBase);

	return Skew;
}
