#include "Bitmap.h"
#include "SDL_image.h"
#include <string.h>

GLuint CBitmap::ActiveTexId;

CBitmap::CBitmap()
{
	ImageData = NULL;
	TexId = 0;
	BitmapSet->Register(this);
	Mask = NULL;
}

CBitmap::~CBitmap()
{
	delete[] ImageData;
	delete[] Mask;
	glDeleteTextures(1, &TexId);
	BitmapSet->DeRegister(this);
}

bool CBitmap::PointOver(float x, float y)
{
	if(!Mask) return false;
	int iy, ix;
	
	iy = (int)(0.5f + y*ImageHeight);
	ix = (int)(0.5f + x*ImageWidth);
	
	return Mask[(iy*MaskPitch) + (ix >> 5)]&(1 << (ix&31)) ? true : false;
}

Uint32 mySDL_GetColour(SDL_Surface *Map, int x, int y)
{
	Uint32 SrcColour = 0;
	Uint8 *LineBase = &((Uint8 *)Map->pixels)[y*Map->pitch];
	memcpy(&SrcColour, &LineBase[x * Map->format->BytesPerPixel], Map->format->BytesPerPixel);
	return SrcColour;
}

void CBitmap::Draw(float x1, float y1, float x2, float y2, int flags)
{
	Activate();
	float EndW = (TexTarget == GL_TEXTURE_2D) ? 1 : ImageWidth;
	float EndH = (TexTarget == GL_TEXTURE_2D) ? 1 : ImageHeight;
	
	glBegin(GL_QUADS);
		glTexCoord2f((flags&BMP_HFLIP) ? EndW : 0, (flags&BMP_VFLIP) ? EndH : 0); glVertex2f(x1, y1);
		glTexCoord2f((flags&BMP_HFLIP) ? 0 : EndW, (flags&BMP_VFLIP) ? EndH : 0); glVertex2f(x2, y1);
		glTexCoord2f((flags&BMP_HFLIP) ? 0 : EndW, (flags&BMP_VFLIP) ? 0 : EndH); glVertex2f(x2, y2);
		glTexCoord2f((flags&BMP_HFLIP) ? EndW : 0, (flags&BMP_VFLIP) ? 0 : EndH); glVertex2f(x1, y2);
	glEnd();
}

bool CBitmap::Load(char *name, bool s)
{
	SDL_Surface *Image = IMG_Load(name);
	if(!Image)
		return false;

	ImageWidth = Image->w;
	ImageHeight = Image->h;
	int WP2, HP2, TargetX = 0, TargetY = 0;
	WP2 = 1;
	while(WP2 < ImageWidth) WP2 <<= 1;
	HP2 = 1;
	while(HP2 < ImageHeight) HP2 <<= 1;
	
	if(HP2 == ImageHeight && WP2 == ImageWidth)
	{
		TexTarget = GL_TEXTURE_2D;
	}
	else
	{
		if(!(GLCapabilities&GLCAP_TEXTURE_RECTANGLE))
		{
			TargetX = (WP2 - ImageWidth) >> 1;
			TargetY = HP2 - ImageHeight;

			ImageWidth = WP2;
			ImageHeight = HP2;

			TexTarget = GL_TEXTURE_2D;
		}
		else
			TexTarget = GLRectangleTexTarget;
	}

	ImageData = new unsigned char[ImageWidth*ImageHeight*4];
	memset(ImageData, 0, ImageWidth*ImageHeight*4);
	SmoothScale = s;

	/* prepare mask data */
	delete[] Mask;
	MaskPitch = (ImageWidth+31) >> 5;
	Mask = new Uint32[MaskPitch*ImageHeight];
	memset(Mask, 0, MaskPitch*ImageHeight*sizeof(Uint32));

	for(int y = 0; y < Image->h; y++)
	{
		for(int x = 0; x < Image->w; x++)
		{
			Uint8 r, g, b, a;
			SDL_GetRGBA(mySDL_GetColour(Image, x, y), Image->format, &r, &g, &b, &a);

			if(r == 0xff && g == 0 && b == 0xff && (Image->format->BytesPerPixel < 4)) a = 0;

			int tx = x+TargetX;
			int ty = y+TargetY;

			int basec = (ty*4*ImageWidth) + tx*4;
			ImageData[basec + 0] = r;
			ImageData[basec + 1] = g;
			ImageData[basec + 2] = b;
			ImageData[basec + 3] = a;

			if(a > 127) Mask[(ty*MaskPitch) + (tx >> 5)] |= 1 << (tx&31);
		}
	}
	Restore();

	SDL_FreeSurface(Image);
	return false;
}

int TexMode = 0;

void DisableTextures()
{
	if(TexMode)
	{
		glDisable(GL_TEXTURE_2D);
		glDisable(GLRectangleTexTarget);
	}
	TexMode = 0;
}

void CBitmap::Activate()
{
	if(TexMode != TexTarget)
	{
		if(TexMode) glDisable(TexMode);
		glEnable(TexMode = TexTarget);
	}
	if(TexId != ActiveTexId)
	{
		ActiveTexId = TexId;
		glBindTexture(TexTarget, TexId);
	}
}

void CBitmap::Backup()
{
	if(TexId)
	{
		ImageData = new unsigned char[ImageWidth*ImageHeight*4];
		glBindTexture(TexTarget, TexId);
		glGetTexImage(TexTarget, 0, GL_RGBA, GL_UNSIGNED_BYTE, ImageData);
		glDeleteTextures(1, &TexId);
		ActiveTexId = 0;
	}
}

void CBitmap::Restore()
{
	if(ImageData)
	{
		glGenTextures(1, &TexId);
		glBindTexture(TexTarget, TexId);

		glTexParameteri(TexTarget, GL_TEXTURE_WRAP_S, SmoothScale ? GL_CLAMP_TO_EDGE : GL_CLAMP);
		glTexParameteri(TexTarget, GL_TEXTURE_WRAP_T, SmoothScale ? GL_CLAMP_TO_EDGE : GL_CLAMP);
		glTexParameteri(TexTarget, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(TexTarget, GL_TEXTURE_MAG_FILTER, SmoothScale ? GL_LINEAR : GL_NEAREST);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);

		glTexImage2D(TexTarget, 0, GL_RGBA, ImageWidth, ImageHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, ImageData);

		delete[] ImageData;
		ImageData = NULL;
	}
}

/*

	CBITMAPSET

*/

CBitmapSet *BitmapSet;

CBitmapSet::CBitmapSet()
{
	Head = NULL;
}

CBitmapSet::~CBitmapSet()
{
	while(Head)
	{
		BitmapPtr *Next = Head->Next;
		delete Head->Data;
		delete Head;
		Head = Next;
	}
}

void CBitmapSet::Backup()
{
	BitmapPtr *p = Head;
	while(p)
	{
		p->Data->Backup();
		p = p->Next;
	}
}

void CBitmapSet::Restore()
{
	BitmapPtr *p = Head;
	while(p)
	{
		p->Data->Restore();
		p = p->Next;
	}
}

void CBitmapSet::Register(CBitmap *n)
{
	BitmapPtr *NewPtr = new BitmapPtr;
	NewPtr->Data = n;
	NewPtr->Next = Head;
	Head = NewPtr;
}

void CBitmapSet::DeRegister(CBitmap *n)
{
	BitmapPtr **Bpt = &Head;
	while(*Bpt)
	{
		if((*Bpt)->Data == n)
		{
			BitmapPtr *Del = *Bpt;
			*Bpt = (*Bpt)->Next;
			delete Del;
			return;
		}
		Bpt = &(*Bpt)->Next;
	}
}
