#include "TrackEdit.h"
#include <stdio.h>
#include <math.h>
#include "3dView.h"

CElevationEditor::CElevationEditor()
{
	DragElevation = NULL;
	DragCorner = NULL;
}

void CElevationEditor::SetPos(float x1, float y1, float x2, float y2)
{
	CWidgetBase::SetPos(x1, y1, x2, y2);
	CentreY = (y1+y2)*0.5f;
}

void CElevationEditor::DoCurve(ElevationNode *a, ElevationNode *b)
{
	/* quadratic bezier */
	glColor3f(1, 1, 1);
	glBegin(GL_LINE_STRIP);
	float t = 0;
	while(t <= 1)
	{
		float fx, fy;
		EvaluateElevation(a, b, t, &fx, &fy);
		glVertex2f((fx - StartY)*TRACK_SCALE + x1, CentreY + fy*TRACK_SCALE);

		t += 0.01;
	}
	glEnd();
}

void CElevationEditor::StartCurve(ElevationNode *a, ElevationNode *b)
{
	CurvePos = 0;
	EvaluateElevation(a, b, 0, &x[1], &y[1]);
}

void CElevationEditor::AdvanceCurve(ElevationNode *a, ElevationNode *b)
{
	x[0] = x[1]; y[0] = y[1];
	CurvePos++;
	EvaluateElevation(a, b, (float)CurvePos / 100.0f, &x[1], &y[1]);
	SectionLength = sqrt((x[1]-x[0])*(x[1]-x[0]) + (y[1]-y[0])*(y[1]-y[0]));
	CurLength += SectionLength;
}

ElevationNode *CElevationEditor::GetElevationNode(float x, float y, int &dir)
{
	float GrabSize = 0.01f/TRACK_SCALE;
	float xpos = ((x - x1) / TRACK_SCALE) + StartY;
	float ypos = (y - CentreY) / TRACK_SCALE;

	/* search for a draggable node */
	int P = (int)(StartY/SEGMENT_LENGTH);
	Segment = &Paths[SectionIndices[P]];
	ElevationNode *Search = Paths[SectionIndices[P]].ElevationList;
	while(Search && Search->y+GrabSize < xpos)
		Search = Search->Next;

	if(Search && fabs(Search->y - xpos) < GrabSize && fabs(Search->z - ypos) < GrabSize)
	{
		dir = 0;
		return Search;
	}

	/* search for a draggable end point */
	Search = Paths[SectionIndices[P]].ElevationList;
	while(Search)
	{
		if(Search->y+Search->Tangent[0]+GrabSize >= xpos) break;
		Search = Search->Next;
	}

	if(Search && fabs(Search->y+Search->Tangent[0] - xpos) < GrabSize && fabs(Search->z+Search->Tangent[1] - ypos) < GrabSize)
	{
		dir = 1;
		return Search;
	}

	return NULL;
}

CornerNode *CElevationEditor::GetCornerNode(float x, float y, int &dir)
{
	float GrabSize = 0.01f;
	int P = (int)(StartY/SEGMENT_LENGTH);
	Segment = &Paths[SectionIndices[P]];

	/* search first for curve controllers */
	float RadiusSq = (y2-y1)*0.04f;
	RadiusSq *= RadiusSq;
	CornerNode *Search = Paths[SectionIndices[P]].CornerList;
	while(Search && Search->y < ScrollBase)
		Search = Search->Next;
	CornerNode *Start = Search;

	while(Search)
	{
		float EndX, EndY;
		EndX = Search->scry + (y2-y1)*0.2f*sin(Search->Curvature);
		EndY = y1 + (y2-y1)*0.2f*cos(Search->Curvature);

		if(((x - EndX)*(x - EndX) + (y - EndY)*(y - EndY)) < RadiusSq)
		{
			dir = 1;
			return Search;
		}

		Search = Search->Next;
	}

	/* failing that search for corner positions */
	Search = Start;	
	while(Search && Search->scry+GrabSize < x)
		Search = Search->Next;
	if(Search && Search->scry-GrabSize < x)
	{
		dir = 0;
		return Search;
	}

	return NULL;
}

float CElevationEditor::GetArcPos(float y)
{
	y = (y - x1)/TRACK_SCALE + StartY;

	/* find node that y rests within */
	ElevationNode *ParseNode = Paths[SectionIndices[ScrollBase/SEGMENT_LENGTH]].ElevationList;
	while(ParseNode->Next && ParseNode->Next->y < y)
		ParseNode = ParseNode->Next;

	if(!ParseNode->Next) return 0;

	/* find segment that y rests within */
	CurLength = ParseNode->ArcStart;
	StartCurve(ParseNode, ParseNode->Next);

	do
		AdvanceCurve(ParseNode, ParseNode->Next);
	while(x[1] < y);

	/* determine mapped y */
	float Ratio = (x[1] - y) / (x[1] - x[0]);
	return CurLength - Ratio*SectionLength;
}

void CElevationEditor::DrawCornerMark(float x, CornerNode *Src)
{
	Src->scry = x;

	glBegin(GL_QUADS);
		glColor4f(1, 1, 0, 0.25f);
		glVertex2f(x-0.01f, y1); glVertex2f(x+0.01f, y1);
		glVertex2f(x+0.01f, y2); glVertex2f(x-0.01f, y2);
		
		float EndX, EndY, SideX, SideY;
		
		glColor4f(1, 0, 0, 0.5f);
		EndX = x + (y2-y1)*0.2f*sin(Src->Curvature);
		EndY = y1 + (y2-y1)*0.2f*cos(Src->Curvature);
		SideX = (y2-y1)*0.02f*cos(Src->Curvature);
		SideY = -(y2-y1)*0.02f*sin(Src->Curvature);
		
		glVertex2f(x-SideX, y1-SideY); glVertex2f(x+SideX, y1+SideY);
		glVertex2f(EndX+SideX, EndY+SideY); glVertex2f(EndX-SideX, EndY-SideY);
	glEnd();
}

bool CElevationEditor::Message(DialogueMessage *Msg)
{
	CDialogueBase::Message(Msg);

	switch(Msg->Type)
	{
		default: break;
		case DialogueMessage::KEYPRESS:
			if((Msg->Data.Key >> 16) == SDLK_SPACE &&
				Msg->x >= x1 && Msg->x <= x2 && Msg->y >= y1 && Msg->y <= y2)
			{
				float ArcPos = GetArcPos(Msg->x);
				if(ArcPos)
				{
					Paths[SectionIndices[(int)(StartY/SEGMENT_LENGTH)]].AddCornerNodeAt(ArcPos);
					SceneViewer->FixCorners(SectionIndices[(int)(StartY/SEGMENT_LENGTH)]);
					SceneViewer->FixObjects(SectionIndices[(int)(StartY/SEGMENT_LENGTH)]);
				}
			}
		break;
		case DialogueMessage::CLICK:
			if(Msg->x < x1) return false;
			if(Msg->x > x2) return false;
			if(Msg->y < y1) return false;
			if(Msg->y > y2) return false;

			if(Msg->Buttons&4) /* try for a split */
			{
				float pos = ((Msg->x - x1) / TRACK_SCALE) + StartY;
				int Tile = (int)pos / SEGMENT_LENGTH;
				if(Tile < 8)
				{
					Paths[SectionIndices[Tile]].SplitElevationAt(pos);
					SceneViewer->FixElevation(SectionIndices[Tile]);
					SceneViewer->FixObjects(SectionIndices[Tile]);

					return true;
				}
			}

			if(Msg->Data.ClickChange&1) /* try for a grab */
			{
				if(Msg->Buttons&1)
				{
					if(DragElevation = GetElevationNode(Msg->x, Msg->y, DragTarget)) return true;
					if(DragCorner = GetCornerNode(Msg->x, Msg->y, DragTarget)) return true;
					
					return false;
				}
				else
				{
					DragElevation = NULL;
					DragCorner = NULL;
					return true;
				}
			}
			if(Msg->Buttons&Msg->Data.ClickChange&2) /* check for a delete */
			{
				int Dir;
				ElevationNode *Elim = GetElevationNode(Msg->x, Msg->y, Dir);
				if(Elim && !Dir)
				{
					if(Elim->Next && Elim != Paths[Segment - Paths].ElevationList)
					{
						ElevationNode *Before = Segment->ElevationList;
						while(Before->Next != Elim)
							Before = Before->Next;
						Before->Next = Elim->Next;
						delete Elim;

						SceneViewer->FixElevation(Segment - Paths);
						SceneViewer->FixObjects(Segment - Paths);
					}
				}
				CornerNode *ElimC = GetCornerNode(Msg->x, Msg->y, Dir);
				if(ElimC && !Dir)
				{
					CornerNode **Place = &Segment->CornerList;
					while(*Place != ElimC)
						Place = &(*Place)->Next;
					*Place = (*Place)->Next;
					delete ElimC;

					SceneViewer->FixCorners(Segment - Paths);
					SceneViewer->FixObjects(Segment - Paths);
				}
			}
		break;
		case DialogueMessage::MOUSEMOVE:
			if(Msg->x >= x1 && Msg->x <= x2 && Msg->y >= y1 && Msg->y <= y2)
			{
				MouseYPos = GetArcPos(Msg->x);

				float z = (Msg->y - CentreY) / TRACK_SCALE;
				float y = ((Msg->x - x1) / TRACK_SCALE) + StartY;
				if(DragElevation)
				{
					switch(DragTarget)
					{
						default:
							if(DragElevation->Next)
							{
								DragElevation->Tangent[0] += DragElevation->y - y;
								DragElevation->y = y;
								DragElevation->Tangent[1] += DragElevation->z - z;
							}

							DragElevation->z = z;
						break;
						case 1:
							DragElevation->Tangent[0] = y - DragElevation->y;
							DragElevation->Tangent[1] = z - DragElevation->z;
						break;
						case -1:
							DragElevation->Tangent[0] = DragElevation->y - y;
							DragElevation->Tangent[1] = DragElevation->z - z;
						break;
					}
					Segment->FixElevation();
					SceneViewer->FixElevation(Segment - Paths);
					SceneViewer->FixObjects(Segment - Paths);
				}
				if(DragCorner)
				{
					switch(DragTarget)
					{
						default:
							DragCorner->y = GetArcPos(Msg->x);
						break;
						case 1:
							DragCorner->Curvature = atan2(Msg->x - DragCorner->scry, Msg->y - y1);
							if(DragCorner->Curvature < -1) DragCorner->Curvature = -1;
							if(DragCorner->Curvature > 1) DragCorner->Curvature = 1;
						break;
					}
					Segment->FixCorners();
					SceneViewer->FixCorners(Segment - Paths);
					SceneViewer->FixObjects(Segment - Paths);
				}
			}
		return false;

		case DialogueMessage::DRAW:
			GLFoundation.SetScissor(x1, y1, x2, y2);

			DisableTextures();
			glColor3f(1, 1, 1);

			/* find first segment that has a visible portion */
				ElevationNode *ParseNode = Paths[SectionIndices[ScrollBase/SEGMENT_LENGTH]].ElevationList;
				while(ParseNode->Next && ParseNode->Next->ArcStart < ScrollBase)
					ParseNode = ParseNode->Next;

			/* find out where visible portion begins */
				CurLength = ParseNode->ArcStart;
				StartCurve(ParseNode, ParseNode->Next);

				do
					AdvanceCurve(ParseNode, ParseNode->Next);
				while(CurLength < ScrollBase);

			/* seed scroll variable */
				StartY = x[0] + (1.0f - ((CurLength - ScrollBase)/SectionLength))*(x[1] - x[0]);

			/* find first visible corner */
				CornerNode *CurCorner = Paths[SectionIndices[ScrollBase/SEGMENT_LENGTH]].CornerList;
				while(CurCorner && CurCorner->y < ScrollBase)
					CurCorner = CurCorner->Next;

			/* draw to end of visibility */
			while(1)
			{
				while(1)
				{
					glBegin(GL_LINES);
						glVertex2f((x[0] - StartY)*TRACK_SCALE + x1, CentreY + y[0]*TRACK_SCALE);
						glVertex2f((x[1] - StartY)*TRACK_SCALE + x1, CentreY + y[1]*TRACK_SCALE);

					/* has a 16 place boundary, the MouseYPos or any corner mark been passed? */
					int S, E;

					E = (int)(CurLength / 16.0f);
					S = (int)((CurLength - SectionLength) / 16.0f);
					if(S != E)
					{
						glColor4f(0, 0, 1, 0.5);

						float sub = fmod(CurLength, 16.0f);
						float SY = x[0] + (1.0f - (sub/SectionLength))*(x[1] - x[0]);
						float xpos = (SY - StartY)*TRACK_SCALE + x1;

						glVertex2f(xpos, y1);
						glVertex2f(xpos, y2);
						glColor3f(1, 1, 1);
					}

					if(CurLength - SectionLength <= MouseYPos && CurLength >= MouseYPos)
					{
						float Ratio = (MouseYPos - CurLength + SectionLength) / SectionLength;
						float xpos = x[0] + Ratio*(x[1] - x[0]);
						float px = (xpos - StartY)*TRACK_SCALE + x1;

						glColor4f(1, 0, 1, 0.2f);
						glVertex2f(px, y1);
						glVertex2f(px, y2);
						glColor3f(1, 1, 1);
					}

					glEnd();
					while(CurCorner && CurLength - SectionLength <= CurCorner->y && CurLength >= CurCorner->y)
					{
						float Ratio = (CurCorner->y - CurLength + SectionLength) / SectionLength;
						float xpos = x[0] + Ratio*(x[1] - x[0]);
						float px = (xpos - StartY)*TRACK_SCALE + x1;

						DrawCornerMark(px, CurCorner);
						glColor3f(1, 1, 1);
						CurCorner = CurCorner->Next;
					}

					if(CurvePos == 100)
						break;

					if(((x[1] - StartY)*TRACK_SCALE + x1) > x2)
						break;
					AdvanceCurve(ParseNode, ParseNode->Next);
				}
				ParseNode = ParseNode->Next;

				if(	((x[1] - StartY)*TRACK_SCALE + x1) > x2 ||
					!ParseNode->Next) break;
				StartCurve(ParseNode, ParseNode->Next);
			}

			/* add cross and tangent */
			ParseNode = Paths[SectionIndices[ScrollBase/SEGMENT_LENGTH]].ElevationList;
			while(ParseNode->Next && ParseNode->Next->y < StartY)
				ParseNode = ParseNode->Next;

			while(ParseNode)
			{
				float fx = (ParseNode->y - StartY)*TRACK_SCALE + x1;
				float fy = CentreY + ParseNode->z*TRACK_SCALE;
				float vx = ParseNode->Tangent[0]*TRACK_SCALE;
				float vy = ParseNode->Tangent[1]*TRACK_SCALE;
				
				if(fx > x2) break;

				glBegin(GL_LINES);
					glColor4f(1, 0, 0, 0.5);
					glVertex2f(fx-0.01f, fy-0.01f); glVertex2f(fx+0.01f, fy+0.01f);
					glVertex2f(fx-0.01f, fy+0.01f); glVertex2f(fx+0.01f, fy-0.01f);

					glVertex2f(fx+vx-0.01f, fy+vy-0.01f); glVertex2f(fx+vx+0.01f, fy+vy+0.01f);
					glVertex2f(fx+vx-0.01f, fy+vy+0.01f); glVertex2f(fx+vx+0.01f, fy+vy-0.01f);
					glColor4f(0, 1, 0, 0.5);
					glVertex2f(fx, fy); glVertex2f(fx+vx, fy+vy);
				glEnd();

				ParseNode = ParseNode->Next;
			}

			GLFoundation.DisableScissor();
		return false;
	}

	return false;
}
