#include "MapSegments.h"
#include <stdio.h>
#include "ImageArray.h"
#include <math.h>

/*

	MAPSEGMENT

*/

MapSegment::MapSegment()
{
	ObjList = NULL;
	ElevationList = NULL;
	CornerList = NULL;
}

MapSegment::~MapSegment()
{
	while(ObjList)
	{
		Object *NPtr = ObjList->Next;
		delete ObjList;
		ObjList = NPtr;
	}

	while(ElevationList)
	{
		ElevationNode *NPtr = ElevationList->Next;
		delete ElevationList;
		ElevationList = NPtr;
	}

	while(CornerList)
	{
		CornerNode *NPtr = CornerList->Next;
		delete CornerList;
		CornerList = NPtr;
	}
}

void MapSegment::AddObject(Object *Obj)
{
	/* find correct position to insert */
	Object **InsertPoint = &ObjList;
	while(*InsertPoint && (*InsertPoint)->y > Obj->y)
		InsertPoint = &(*InsertPoint)->Next;

	Obj->Next = *InsertPoint;
	*InsertPoint = Obj;
}

bool MapSegment::RemoveObject(Object *Obj)
{
	Object **InsertPoint = &ObjList;
	while(*InsertPoint && (*InsertPoint) != Obj)
		InsertPoint = &(*InsertPoint)->Next;

	if(*InsertPoint)
	{
		*InsertPoint = Obj->Next;
		delete Obj;
		return true;
	}
	return false;
}

void MapSegment::AddCornerNodeAt(float y)
{
	CornerNode **InsertPoint = &CornerList;
	while(*InsertPoint && (*InsertPoint)->y < y)
		InsertPoint = &(*InsertPoint)->Next;

	CornerNode *CN = new CornerNode;
	CN->y = y;
	CN->Curvature = 0;
	CN->Next = *InsertPoint;
	*InsertPoint = CN;
}

void MapSegment::FixCorners()
{
	CornerNode *P = CornerList;
	
	if(P)
	{
		if(P->y < StartY+1) P->y = StartY+1;
		while(P->Next)
		{
			if(P->y > P->Next->y) P->Next->y = P->y;
			P = P->Next;
		}
	}
}

void EvaluateElevation(ElevationNode *a, ElevationNode *b, float t, float *y, float *z, float *tangent)
{
	float bx = (a->y + a->Tangent[0]) + t*(b->y - b->Tangent[0] - a->y - a->Tangent[0]);
	float by = (a->z + a->Tangent[1]) + t*(b->z - b->Tangent[1] - a->z - a->Tangent[1]);

	float dx, dy, ex, ey;
	dx = (1-t)*(a->y + t*a->Tangent[0]) + t*bx;
	dy = (1-t)*(a->z + t*a->Tangent[1]) + t*by; 
	ex = (1-t)*bx + t*(b->y - b->Tangent[0] + t*b->Tangent[0]);
	ey = (1-t)*by + t*(b->z - b->Tangent[1] + t*b->Tangent[1]);

	if(y) *y = dx + t*(ex - dx);
	if(z) *z = dy + t*(ey - dy);

	if(tangent)
	{
		tangent[0] = (ex - dx)*0.25f;
		tangent[1] = (ey - dy)*0.25f;
	}
}

void MapSegment::SplitElevationAt(float iny)
{
	/* do a binary search for t */
	ElevationNode *S = ElevationList;
	while(S->Next && S->Next->y < iny)
		S = S->Next;

	/* binary search */
	float tl = 0;
	float th = 1;
	float y, z, Tangent[2];
	while(1)
	{
		EvaluateElevation(S, S->Next, (tl+th)*0.5f, &y, &z, Tangent);

		if(fabs(y-iny) < 0.1f) break;
		if(y > iny)
			th = (tl+th)*0.5f;
		if(y < iny)
			tl = (tl+th)*0.5f;
	}

	ElevationNode *N = new ElevationNode;
	N->y = y;
	N->z = z;
	N->Tangent[0] = Tangent[0];
	N->Tangent[1] = Tangent[1];
	float Len = sqrt(Tangent[0]*Tangent[0] + Tangent[1]*Tangent[1]);
	if(Len > 5.0f)
	{
		Len = 5.0f / Len;
		N->Tangent[0] *= Len;
		N->Tangent[1] *= Len;
	}
	
	N->Next = S->Next;
	S->Next = N;

	FixElevation();
}

float MapSegment::GetArcLength()
{
	float Length = 0;
	ElevationNode *P = ElevationList;
	while(P->Next)
	{
		float x[2], y[2];
		EvaluateElevation(P, P->Next, 0, &x[1], &y[1]);

		P->ArcStart = Length + StartY;
		for(int step = 0; step < 100; step++)
		{
			float t = (float)(step+1) / 100.0f;
			EvaluateElevation(P, P->Next, t, x, y);
			Length += sqrt((x[1]-x[0])*(x[1]-x[0]) + (y[1]-y[0])*(y[1]-y[0]));
			x[1] = x[0]; y[1] = y[0];
		}
		P = P->Next;
	}
	P->ArcStart = Length + StartY;
	return Length;
}

void MapSegment::FixElevation()
{
	/* apply constraints - first the easy */
	ElevationList->y = StartY;
	ElevationList->Tangent[1] = 0;
	if(ElevationList->Tangent[0] < 1) ElevationList->Tangent[0] = 1;
	
	ElevationNode *P = ElevationList;
	while(P->Next)
		P = P->Next;
	P->Tangent[1] = 0;
	if(P->Tangent[0] < 1) P->Tangent[0] = 1;

	P = ElevationList;
	while(P)
	{
		if(P->Next)
		{
			if(P->Next->y < P->y) P->Next->y = P->y;
		}
		if(P->Tangent[0] < 0) P->Tangent[0] = 0;
		P = P->Next;
	}

	/* now arc length */
	float Length = GetArcLength();

	float Mul = (float)SEGMENT_LENGTH / Length;
	P = ElevationList;
	while(P)
	{
		P->y = StartY + (P->y - StartY)*Mul;
		P->z *= Mul;
		P->Tangent[0] *= Mul;
		P->Tangent[1] *= Mul;
		P = P->Next;
	}
}

/*

	MISCELLANEOUS - loading, saving

*/

MapSegment Paths[26];
char Filename[2048];
void SetFilename(int id)
{
	/* determine correct path here maybe ? */
	sprintf(Filename, "section%d.dat", id);
}

#define	Get32(v)	\
	v = fgetc(input);\
	v |= fgetc(input) << 8;\
	v |= fgetc(input) << 16;\
	v |= fgetc(input) << 24;

#define Put32(v)	\
	fputc(v&0xff, input);\
	fputc((v >> 8)&0xff, input);\
	fputc((v >> 16)&0xff, input);\
	fputc((v >> 24)&0xff, input);

void LoadSegments()
{
	/* stuff that doesn't need loading */
	Paths[0].Next[0] = 1;
	Paths[0].Next[1] = 2;
	Paths[0].StartY = 0;
	Paths[0].Brother = -1;

	Paths[1].Brother = 2;
	Paths[2].Brother = 1;

	Paths[1].Next[0] = 3;
	Paths[1].Next[1] = 4;
	Paths[2].Next[0] = 5;
	Paths[2].Next[1] = 6;
	Paths[1].StartY =
	Paths[2].StartY = SEGMENT_LENGTH;

	Paths[3].Brother = 4;
	Paths[4].Brother = 3;
	Paths[5].Brother = 6;
	Paths[6].Brother = 5;

	Paths[3].Next[0] = 7;
	Paths[3].Next[1] = 8;
	Paths[4].Next[0] = 9;
	Paths[4].Next[1] = 10;
	Paths[5].Next[0] = 9;
	Paths[5].Next[1] = 10;
	Paths[6].Next[0] = 11;
	Paths[6].Next[1] = 12;
	Paths[3].StartY =
	Paths[4].StartY =
	Paths[5].StartY =
	Paths[6].StartY = 2*SEGMENT_LENGTH;

	Paths[7].Brother = 8;
	Paths[8].Brother = 7;
	Paths[9].Brother = 10;
	Paths[10].Brother = 9;
	Paths[11].Brother = 12;
	Paths[12].Brother = 11;

	Paths[7].Next[0] = -1;
	Paths[7].Next[1] = 18;
	Paths[8].Next[0] = 17;
	Paths[8].Next[1] = 16;
	Paths[9].Next[0] = 17;
	Paths[9].Next[1] = 16;
	Paths[10].Next[0] = 15;
	Paths[10].Next[1] = 14;
	Paths[11].Next[0] = 15;
	Paths[11].Next[1] = 14;
	Paths[12].Next[0] = 13;
	Paths[12].Next[1] = -1;
	Paths[7].StartY =
	Paths[8].StartY =
	Paths[9].StartY =
	Paths[10].StartY =
	Paths[11].StartY =
	Paths[12].StartY = 3*SEGMENT_LENGTH;

	Paths[18].Brother = -1;
	Paths[17].Brother = 16;
	Paths[16].Brother = 17;
	Paths[15].Brother = 14;
	Paths[14].Brother = 15;
	Paths[13].Brother = -1;

	Paths[18].Next[0] = -1;
	Paths[18].Next[1] = 22;
	Paths[17].Next[0] = -1;
	Paths[17].Next[1] = 22;
	Paths[16].Next[0] = 21;
	Paths[16].Next[1] = 20;
	Paths[15].Next[0] = 21;
	Paths[15].Next[1] = 20;
	Paths[14].Next[0] = 19;
	Paths[14].Next[1] = -1;
	Paths[13].Next[0] = 19;
	Paths[13].Next[1] = -1;
	Paths[13].StartY =
	Paths[14].StartY =
	Paths[15].StartY =
	Paths[16].StartY =
	Paths[17].StartY =
	Paths[18].StartY = 4*SEGMENT_LENGTH;

	Paths[22].Brother = -1;
	Paths[21].Brother = 20;
	Paths[20].Brother = 21;
	Paths[19].Brother = -1;

	Paths[22].Next[0] = -1;
	Paths[22].Next[1] = 24;
	Paths[21].Next[0] = -1;
	Paths[21].Next[1] = 24;
	Paths[20].Next[0] = 23;
	Paths[20].Next[1] = -1;
	Paths[19].Next[0] = 23;
	Paths[19].Next[1] = -1;
	Paths[19].StartY =
	Paths[20].StartY =
	Paths[21].StartY =
	Paths[22].StartY = 5*SEGMENT_LENGTH;

	Paths[24].Brother = -1;
	Paths[23].Brother = -1;

	Paths[24].Next[0] = -1;
	Paths[24].Next[1] = 25;
	Paths[23].Next[0] = 25;
	Paths[23].Next[1] = -1;
	Paths[23].StartY =
	Paths[24].StartY = 6*SEGMENT_LENGTH;

	Paths[25].Brother = -1;

	Paths[25].Next[0] = -1;
	Paths[25].Next[1] = -1;
	Paths[25].StartY = 7*SEGMENT_LENGTH;

	/* stuff that does need loading */
	int c = 26;
	while(c--)
	{
		Paths[c].Selected = false;
		if(!c || c == 25) Paths[c].Selected = true;
		sprintf(Paths[c].Name, "Section %d", c);

		SetFilename(c);
		FILE *input;
		if(input = fopen(Filename, "rb"))
		{
			/* read name */
			fread(Paths[c].Name, 1, 33, input);

			/* read flags */
			unsigned char flags;
			flags = fgetc(input);
			Paths[c].Selected = (flags&1) ? true : false;

			/* read tilemap */
			for(int y = 0; y < SEGMENT_LENGTH; y++)
				for(int x = 0; x < 8; x++)
				{
					Paths[c].MaterialMap[y][x] = fgetc(input);
					Paths[c].MaterialMap[y][x] |= fgetc(input) << 8;

					if((Paths[c].MaterialMap[y][x]&TILE_MASK) >= MaterialArray->GetNumImages())
						Paths[c].MaterialMap[y][x] = 0xffff;
				}

			/* read objects */
			int ic = fgetc(input);
			ic |= fgetc(input) << 8;

			if(!feof(input))
			{
				Object **InsertPoint = &Paths[c].ObjList;
				while(ic--)
				{
					int IX, IY;

					Get32(IX);
					Get32(IY);

					*InsertPoint = new Object;
					(*InsertPoint)->y = (float)IY / 256.0f;
					(*InsertPoint)->x = (float)IX / 256.0f;

					(*InsertPoint)->ImageId = fgetc(input);
					(*InsertPoint)->ImageId |= fgetc(input) << 8;

					if(((*InsertPoint)->ImageId&TILE_MASK) >= ObjectArray->GetNumImages())
					{
						delete *InsertPoint;
						*InsertPoint = NULL;
					}
					else
					{
						(*InsertPoint)->Next = NULL;
						InsertPoint = &(*InsertPoint)->Next;
					}
				}
			}

			/* read elevation changes */
			ic = fgetc(input);
			ic |= fgetc(input) << 8;

			if(!feof(input))
			{
				ElevationNode **InsertPoint = &Paths[c].ElevationList;
				while(ic--)
				{
					int IZ, IY;

					Get32(IY);
					Get32(IZ);

					*InsertPoint = new ElevationNode;
					(*InsertPoint)->y = (float)IY / 256.0f;
					(*InsertPoint)->z = (float)IZ / 65536.0f;

					int TanX, TanY;
					Get32(TanX);
					Get32(TanY);
					(*InsertPoint)->Tangent[0] = (float)TanX / 65536.0f;
					(*InsertPoint)->Tangent[1] = (float)TanY / 65536.0f;

					(*InsertPoint)->Next = NULL;
					InsertPoint = &(*InsertPoint)->Next;
				}
			}

			/* read corners */
			ic = fgetc(input);
			ic |= fgetc(input) << 8;

			if(!feof(input))
			{
				CornerNode **InsertPoint = &Paths[c].CornerList;
				while(ic--)
				{
					int IY, Curve;

					Get32(IY);
					Get32(Curve);

					*InsertPoint = new CornerNode;
					(*InsertPoint)->y = (float)IY / 256.0f;
					(*InsertPoint)->Curvature = (float)Curve / 65536.0f;

					(*InsertPoint)->Next = NULL;
					InsertPoint = &(*InsertPoint)->Next;
				}
			}
			fclose(input);
		}
		else
		{
			for(int y = 0; y < SEGMENT_LENGTH; y++)
				for(int x = 0; x < 8; x++)
					Paths[c].MaterialMap[y][x] = 0xffff;
		}

		if(!Paths[c].ElevationList)
		{
			/* default - start flat, end flat */
			Paths[c].ElevationList = new ElevationNode;
			Paths[c].ElevationList->Next = NULL;
			Paths[c].ElevationList->Tangent[0] = 1;
			Paths[c].ElevationList->Tangent[1] = 0;
			Paths[c].ElevationList->z = 0;
			Paths[c].ElevationList->y = Paths[c].StartY;

			ElevationNode *N2 = Paths[c].ElevationList->Next = new ElevationNode;
			N2->Next = NULL;
			N2->Tangent[0] = 1;
			N2->Tangent[1] = 0;
			N2->z = 0;
			N2->y = Paths[c].StartY + SEGMENT_LENGTH;
		}

/*		if(!Paths[c].CornerList)
		{
			Paths[c].CornerList = new CornerNode;
			Paths[c].CornerList->Next = NULL;
			Paths[c].CornerList->y = Paths[c].StartY+20;
			Paths[c].CornerList->Curvature = 1.0f;
		}*/

		Paths[c].FixElevation();
	}
}

void SaveSegments()
{
	int c = 26;
	while(c--)
	{
		SetFilename(c);
		FILE *input;
		if(input = fopen(Filename, "wb"))
		{
			/* write name */
			fwrite(Paths[c].Name, 1, 33, input);

			/* write flags */
			unsigned char flags;
			flags = Paths[c].Selected ? 1 : 0;
			putc(flags, input);

			/* write tilemap */
			for(int y = 0; y < SEGMENT_LENGTH; y++)
				for(int x = 0; x < 8; x++)
				{
					fputc(Paths[c].MaterialMap[y][x]&0xff, input);
					fputc(Paths[c].MaterialMap[y][x] >> 8, input);
				}

			/* write objects */
			int ic = 0;
			Object *Parser = Paths[c].ObjList;
			while(Parser)
			{
				ic++;
				Parser = Parser->Next;
			}
			fputc(ic&0xff, input);
			fputc((ic>>8)&0xff, input);

			Parser = Paths[c].ObjList;
			while(Parser)
			{
				int IX, IY;

				IX = (int)(Parser->x * 256.0f);
				IY = (int)(Parser->y * 256.0f);

				Put32(IX);
				Put32(IY);

				fputc(Parser->ImageId&0xff, input);
				fputc((Parser->ImageId >> 8)&0xff, input);

				Parser = Parser->Next;
			}

			/* write elevation changes */
			ic = 0;
			ElevationNode *NP = Paths[c].ElevationList;
			while(NP)
			{
				ic++;
				NP = NP->Next;
			}
			fputc(ic&0xff, input);
			putc((ic >> 8)&0xff, input);

			NP = Paths[c].ElevationList;
			while(NP)
			{
				int IZ, IY;

				IY = (int)(NP->y * 256.0f);
				IZ = (int)(NP->z * 65536.0f);

				Put32(IY);
				Put32(IZ);

				int TanX, TanY;
				TanX = (int)(NP->Tangent[0] * 65536.0f);
				TanY = (int)(NP->Tangent[1] * 65536.0f);
				Put32(TanX);
				Put32(TanY);

				NP = NP->Next;
			}

			/* write corners */
			ic = 0;
			CornerNode *NC = Paths[c].CornerList;
			while(NC)
			{
				ic++;
				NC = NC->Next;
			}
			fputc(ic&0xff, input);
			putc((ic >> 8)&0xff, input);

			NC = Paths[c].CornerList;
			while(NC)
			{
				int IY, Curve;

				IY = (int)(NC->y * 256.0f);
				Curve = (int)(NC->Curvature * 65536.0f);

				Put32(IY);
				Put32(Curve);

				NC = NC->Next;
			}

			/* close file */
			fclose(input);
		}
	}
}
